@@ -3,6 +3,11 @@
 /*
  * Nessus system includes 
  */
+#ifdef _CYGWIN_
+#undef _WIN32
+#endif
+
+
 #ifdef _WIN32
 #include "config.w32"
 #else
@@ -58,6 +63,11 @@
 #include <strings.h>
 #endif /* HAVE_STRINGS_H */
 
+
+#ifdef HAVE_LIMITS_H
+#include <limits.h>
+#endif
+
 #ifdef HAVE_NETDB_H
 #include <netdb.h>
 #endif
@@ -135,6 +145,15 @@
 # endif            
 #endif   
 
+#ifdef HAVE_SETJMP_H
+#include <setjmp.h>
+#endif
+
+#ifdef HAVE_SSL
+#include <openssl/ssl.h>
+#include <openssl/x509.h>
+#endif
+
 /*
  * Threads management
  *
@@ -175,6 +194,8 @@
 # define UNFIX(n) ntohs(n)
 #endif
 
+#include <pcap.h>
+
 #include <ntcompat.h>
 #include <libnessus.h>
 #include <nessus-devel.h>
