@@ -1,5 +1,5 @@
 /* Nessuslib -- the Nessus Library
- * Copyright (C) 1998-2000 Renaud Deraison
+ * Copyright (C) 1998 - 2003 Renaud Deraison
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -21,20 +21,27 @@
 #define EXPORTING
 #include <includes.h>
 #include "comm.h"
-
-#ifdef ENABLE_CRYPTO_LAYER
-#include "peks.h"
-#endif
+#include "harglists.h"
+#include "diff.h"
+#include "rand.h"
+#include "services.h"
+#include "store.h"
 
 /* want version stuff */
 #include "libvers.h"
+#include "scanners_utils.h"
 
 
 #undef DEBUG_DIFF_SCAN
 
+
+
 char *nessuslib_version()
 {
-  return strdup(VERSION);
+  static char vers[255];
+  strncpy(vers, VERSION, sizeof(vers) - 1);
+  vers[sizeof(vers) - 1 ] = '\0';
+  return vers;
 }
 ExtFunc
 void nessus_lib_version(major, minor, rev)
@@ -45,15 +52,6 @@ void nessus_lib_version(major, minor, rev)
  *rev   = NL_REV;
 }
 
-#ifdef ENABLE_CRYPTO_LAYER
-ExtFunc
-int nessuslib_cipher_enabled()
-{
- int enabled = 1;
- return(enabled);
-}
-#endif
-
 #ifdef USE_PTHREADS
 ExtFunc 
 int nessuslib_pthreads_enabled()
@@ -64,6 +62,158 @@ int nessuslib_pthreads_enabled()
 #endif
 
 
+
+
+
+/*
+ * Escapes \n and \r properly. The resulting string
+ * is copied in another buffer.
+ */
+ExtFunc char * 
+addslashes(in)
+	char * in;
+{
+ char * ret;
+ char * out = malloc(strlen(in) * 2 + 1);
+ bzero(out, strlen(in) * 2 + 1);
+ ret = out;
+ while(in[0])
+ {
+  if(in[0] == '\\')
+  {
+   out[0] = '\\'; out++;
+   out[0] = '\\'; out++;
+  }
+
+  else if(in[0] == '\n')
+  {
+   out[0] = '\\'; out++;
+   out[0] = 'n'; out++;
+  }
+  else if(in[0] == '\r')
+  {
+    out[0] = '\\'; out++;
+    out[0] = 'r';  out++;
+  }
+  else {
+	  out[0] = in[0];
+	  out++;
+  }
+  in++;
+ }
+ return realloc(ret, strlen(ret) + 1);
+}
+
+/*
+ * Replaces escape codes (\n, \r) by the real value
+ * The resulting string is stored in another buffer
+ */
+ExtFunc char * 
+rmslashes(in)
+ char * in;
+{
+ char * out = malloc(strlen(in) + 1);
+ char * ret = out;
+ bzero(out, strlen(in) + 1);
+ while(in[0])
+ {
+  if(in[0] == '\\')
+  {
+   switch(in[1])
+   {
+    case 'r' :
+      out[0] = '\r';
+      in++;
+      break;
+    case 'n' :
+      out[0] =  '\n';
+      in++;
+      break;
+    case '\\' :
+      out[0] = '\\';
+      in++;
+      break;
+    default :
+      fprintf(stderr, "Unknown escape sequence '\\%c'\n", in[1]);
+   }
+  }
+  else out[0] = in[0];
+  in++;
+  out++;
+ }
+ return realloc(ret, strlen(ret) + 1);
+}
+
+
+ExtFunc
+void plug_set_version(desc, version)
+ struct arglist * desc;
+ const char* version;
+{
+ arg_add_value(desc, "VERSION", ARG_STRING, strlen(version), estrdup((char*)version));
+}
+
+
+ExtFunc 
+char * _plug_get_version(desc)
+ struct arglist * desc;
+{
+ return arg_get_value(desc, "VERSION");
+}
+
+
+ExtFunc 
+char * plug_get_version(struct arglist * desc)
+{
+ return store_fetch_version(desc);
+}
+
+
+ExtFunc
+void plug_set_path(desc, path)
+ struct arglist * desc;
+ const char * path;
+{
+ arg_add_value(desc, "PATH", ARG_STRING, strlen(path), estrdup((char*)path));
+}
+
+ExtFunc 
+char * _plug_get_path(struct arglist * desc)
+{
+ return arg_get_value(desc, "PATH");
+}
+
+ExtFunc 
+char * plug_get_path(desc)
+ struct arglist * desc;
+{
+ return store_fetch_path(desc);
+}
+
+
+
+ExtFunc
+void plug_set_fname(desc, filename)
+ struct arglist * desc;
+ const char * filename;
+{
+ arg_add_value(desc, "FILENAME", ARG_STRING, strlen(filename), estrdup(filename));
+}
+
+ExtFunc 
+char * _plug_get_fname(struct arglist * desc)
+{
+ return arg_get_value(desc, "FILENAME");
+}
+
+ExtFunc 
+char * plug_get_fname(desc)
+ struct arglist * desc;
+{
+ return _plug_get_fname(desc);
+}
+
+
 ExtFunc
 void plug_set_id(desc, id)
  struct arglist * desc;
@@ -72,28 +222,120 @@ void plug_set_id(desc, id)
  arg_add_value(desc, "ID", ARG_INT, sizeof(int), (void*)id);
 }
 
+ExtFunc int
+_plug_get_id(desc)
+ struct arglist * desc;
+{
+ return (int)arg_get_value(desc, "ID");
+}
+
+ExtFunc int
+plug_get_id(struct arglist * desc)
+{
+ return _plug_get_id(desc);	/* Never cached */
+}
+
+
 ExtFunc
 void plug_set_cve_id(desc, id)
  struct arglist * desc;
  char * id;
 {
- arg_add_value(desc, "CVE_ID", ARG_STRING, strlen(id), estrdup(id));
+ char * old = arg_get_value(desc, "CVE_ID");
+ if(old != NULL)
+ {
+  old = erealloc(old, strlen(old) + strlen(id) + 3);
+  strcat(old, ", ");
+  strcat(old, id);
+  arg_set_value(desc, "CVE_ID", strlen(old), old);
+ }
+ else
+  arg_add_value(desc, "CVE_ID", ARG_STRING, strlen(id), estrdup(id));
 }
 
-ExtFunc 
-void plug_set_see_also(desc, url)
+
+ExtFunc char *
+_plug_get_cve_id(desc)
  struct arglist * desc;
- char * url;
 {
- struct arglist * seealso = arg_get_value(desc, "SEE_ALSO");
- if(!seealso)
- {
-  seealso = emalloc(sizeof(*seealso));
-  arg_add_value(desc, "SEE_ALSO",  ARG_ARGLIST, -1, seealso);
- }
- arg_add_value(seealso, "url", ARG_STRING, strlen(url), estrdup(url));
+ return arg_get_value(desc, "CVE_ID");
 }
 
+ExtFunc char * plug_get_cve_id(struct arglist * desc)
+{
+ return store_fetch_cve_id(desc);
+}
+
+
+ExtFunc
+void plug_set_bugtraq_id(desc, id)
+ struct arglist * desc;
+ char * id;
+{
+ char * old = arg_get_value(desc, "BUGTRAQ_ID");
+ if(old != NULL)
+ { 
+  old = erealloc(old, strlen(old) + strlen(id) + 3);
+  strcat(old, ", ");
+  strcat(old, id);
+  arg_set_value(desc, "BUGTRAQ_ID", strlen(old), old);
+ }
+ else
+  arg_add_value(desc, "BUGTRAQ_ID", ARG_STRING, strlen(id), estrdup(id));
+}
+
+ExtFunc char * _plug_get_bugtraq_id(desc)
+ struct arglist * desc;
+{
+ return arg_get_value(desc, "BUGTRAQ_ID");
+}
+
+ExtFunc char * plug_get_bugtraq_id(struct arglist * desc)
+{
+ return store_fetch_bugtraq_id(desc);
+}
+
+
+
+ExtFunc
+void plug_set_xref(desc,name, value)
+ struct arglist * desc;
+ char * name, * value;
+{
+ char * old = arg_get_value(desc, "XREFS");
+ if(old != NULL)
+ { 
+  old = erealloc(old, strlen(old) + strlen(name) + strlen(value) + 4);
+  strcat(old, ", ");
+  strcat(old, name);
+  strcat(old, ":");
+  strcat(old, value);
+  arg_set_value(desc, "XREFS", strlen(old), old);
+ }
+ else 
+  {
+  char * str;
+  
+  str = emalloc(strlen(name) + strlen(value) + 2);
+  strcat(str, name);
+  strcat(str, ":");
+  strcat(str, value);
+  arg_add_value(desc, "XREFS", ARG_STRING, strlen(str), str);
+  }
+}
+
+ExtFunc char * _plug_get_xref(desc)
+ struct arglist * desc;
+{
+ return arg_get_value(desc, "XREFS");
+}
+
+ExtFunc char * plug_get_xref(struct arglist * desc)
+{
+ return store_fetch_xref(desc);
+}
+
+
 
 ExtFunc
 void plug_set_family(desc, family, language)
@@ -122,6 +364,22 @@ void plug_set_family(desc, family, language)
     }
 }
 
+
+ExtFunc
+char * _plug_get_family(desc)
+ struct arglist * desc;
+{
+ return arg_get_value(desc, "FAMILY");
+}
+
+ExtFunc
+char * plug_get_family(desc)
+ struct arglist * desc;
+{
+ return store_fetch_family(desc);
+}
+
+
 ExtFunc
 void plug_require_key(desc, keyname)
  struct arglist * desc;
@@ -136,10 +394,26 @@ void plug_require_key(desc, keyname)
    keys = emalloc(sizeof(struct arglist));
    arg_add_value(desc, "required_keys", ARG_ARGLIST, -1, keys);
   }
-  arg_add_value(keys, "key", ARG_STRING, strlen(keyname), strdup(keyname));
+  arg_add_value(keys, keyname,  ARG_INT, 0, (void*)1);
  }
 }
 
+ExtFunc 
+struct arglist * _plug_get_required_keys(desc)
+ struct arglist * desc;
+{
+ return arg_get_value(desc, "required_keys");
+}
+
+
+ExtFunc 
+struct arglist * plug_get_required_keys(desc)
+ struct arglist * desc;
+{
+ return store_fetch_required_keys(desc);
+}
+
+
 ExtFunc
 void plug_exclude_key(desc, keyname)
  struct arglist * desc;
@@ -154,10 +428,24 @@ void plug_exclude_key(desc, keyname)
    keys = emalloc(sizeof(struct arglist));
    arg_add_value(desc, "excluded_keys", ARG_ARGLIST, -1, keys);
   }
-  arg_add_value(keys, "key", ARG_STRING, strlen(keyname), strdup(keyname));
+  arg_add_value(keys, keyname, ARG_INT, 0, (void*)1);
  }
 }
 
+ExtFunc
+struct arglist * _plug_get_excluded_keys(desc)
+ struct arglist * desc;
+{
+ return arg_get_value(desc, "excluded_keys");
+}
+
+
+ExtFunc
+struct arglist * plug_get_excluded_keys(desc)
+ struct arglist * desc;
+{
+ return store_fetch_excluded_keys(desc);
+}
 
 ExtFunc 
 void plug_require_port(desc, portname)
@@ -165,9 +453,8 @@ void plug_require_port(desc, portname)
  const char * portname;
 {
  struct arglist * ports;
- int iport = atoi(portname);
  
- if(portname)
+ if(portname != NULL)
  {
   ports = arg_get_value(desc, "required_ports");
   if(!ports)
@@ -176,24 +463,33 @@ void plug_require_port(desc, portname)
    arg_add_value(desc, "required_ports", ARG_ARGLIST, -1, ports);
   }
   
-
-  if(iport)
-   arg_add_value(ports, portname, ARG_INT, sizeof(iport), (char*)iport);
-  else 
-   arg_add_value(ports, portname, ARG_STRING, strlen(portname), strdup(portname));
-
+  arg_add_value(ports, portname, ARG_INT, 0, (void*)1);
  }
 }
 
+ExtFunc
+struct arglist * _plug_get_required_ports(desc)
+ struct arglist * desc;
+{
+ return arg_get_value(desc, "required_ports");
+}
+
+ExtFunc
+struct arglist * plug_get_required_ports(desc)
+ struct arglist * desc;
+{
+ return store_fetch_required_ports(desc);
+}
+
+
 ExtFunc 
 void plug_require_udp_port(desc, portname)
  struct arglist * desc;
  const char * portname;
 {
  struct arglist * ports;
- int iport = atoi(portname);
  
- if(portname)
+ if(portname != NULL)
  {
   ports = arg_get_value(desc, "required_udp_ports");
   if(!ports)
@@ -202,14 +498,23 @@ void plug_require_udp_port(desc, portname)
    arg_add_value(desc, "required_udp_ports", ARG_ARGLIST, -1, ports);
   }
   
-
-  if(iport)
-   arg_add_value(ports, portname, ARG_INT, sizeof(iport), (char*)iport);
-  else 
-   arg_add_value(ports, portname, ARG_STRING, strlen(portname), strdup(portname));
-
+   arg_add_value(ports, portname, ARG_INT, 0, (void*)1);
  }
 }
+
+ExtFunc 
+struct arglist * _plug_get_required_udp_ports(desc)
+ struct arglist * desc;
+{
+ return arg_get_value(desc, "required_udp_ports");
+}
+
+ExtFunc 
+struct arglist * plug_get_required_udp_ports(desc)
+ struct arglist * desc;
+{
+ return store_fetch_required_udp_ports(desc);
+}
  
 
 
@@ -226,9 +531,25 @@ void plug_set_dep(desc, depname)
    deps = emalloc(sizeof(struct arglist));
    arg_add_value(desc, "DEPENDENCIES", ARG_ARGLIST, -1, deps);
    }
-  arg_add_value(deps, estrdup(depname), ARG_STRING, 0, "");
+  arg_add_value(deps, depname, ARG_STRING, 0, estrdup(""));
  }
 }
+
+ExtFunc
+struct arglist * _plug_get_deps(desc)
+ struct arglist * desc;
+{
+ return arg_get_value(desc, "DEPENDENCIES");
+}
+
+
+ExtFunc
+struct arglist * plug_get_deps(desc)
+ struct arglist * desc;
+{
+ return store_fetch_dependencies(desc);
+}
+
 ExtFunc
 void plug_set_timeout(desc, timeout)
  struct arglist * desc; 
@@ -240,11 +561,20 @@ void plug_set_timeout(desc, timeout)
 
 
 ExtFunc
-const char * plug_get_family(desc)
+int _plug_get_timeout(desc)
  struct arglist * desc;
 {
- return(arg_get_value(desc, "FAMILY"));
+ return (int)arg_get_value(desc, "TIMEOUT");
 }
+
+
+ExtFunc
+int plug_get_timeout(desc)
+ struct arglist * desc;
+{
+ return store_fetch_timeout(desc);
+}
+
 		
 
 ExtFunc
@@ -263,7 +593,7 @@ ExtFunc
 int plug_get_launch(desc)
  struct arglist * desc;
 {
- return((int)arg_get_value(desc, "ENABLED"));
+ 	return((int)arg_get_value(desc, "ENABLED"));
 }	
 	
 	
@@ -294,6 +624,20 @@ void plug_set_name(desc, name, language)
   }
 }
 
+ExtFunc
+char * _plug_get_name(desc)
+ struct arglist * desc;
+{
+ return arg_get_value(desc, "NAME");
+}
+
+ExtFunc
+char * plug_get_name(desc)
+ struct arglist * desc;
+{
+ return store_fetch_name(desc);
+}
+
 
 ExtFunc
 void plug_set_summary(desc, summary,language)
@@ -322,6 +666,20 @@ void plug_set_summary(desc, summary,language)
   }
 }
 
+ExtFunc
+char * _plug_get_summary(desc)
+ struct arglist * desc;
+{
+ return arg_get_value(desc, "SUMMARY");
+}
+
+ExtFunc
+char * plug_get_summary(desc)
+ struct arglist * desc;
+{
+ return store_fetch_summary(desc);
+}
+
 
 ExtFunc
 void plug_set_description(desc, description,language)
@@ -351,6 +709,21 @@ void plug_set_description(desc, description,language)
 }
 
 
+ExtFunc
+char * _plug_get_description(desc)
+ struct arglist * desc;
+{
+ return arg_get_value(desc, "DESCRIPTION");
+}
+
+ExtFunc
+char * plug_get_description(desc)
+ struct arglist * desc;
+{
+ return store_fetch_description(desc);
+}
+
+
 ExtFunc
 void plug_set_copyright(desc, copyright,language)
  struct arglist * desc;
@@ -378,6 +751,21 @@ void plug_set_copyright(desc, copyright,language)
   }
 }
 
+ExtFunc
+char * _plug_get_copyright(desc)
+ struct arglist * desc;
+{
+ return arg_get_value(desc, "COPYRIGHT");
+}
+
+
+ExtFunc
+char * plug_get_copyright(desc)
+ struct arglist * desc;
+{
+ return store_fetch_copyright(desc);
+}
+
 
 ExtFunc
 void plug_set_category(desc, category)
@@ -387,16 +775,22 @@ void plug_set_category(desc, category)
        arg_add_value(desc, "CATEGORY", ARG_INT, sizeof(int), (void *)category);
 }
 
+ExtFunc
+int _plug_get_category(desc)
+ struct arglist * desc;
+{
+ return (int)arg_get_value(desc, "CATEGORY");	/* We don't cache this one */
+}
 
 ExtFunc
-void plug_set_attacklevel(desc, attacklevel)
+int plug_get_category(desc)
  struct arglist * desc;
- int attacklevel;
 {
-	arg_add_value(desc, "ATTACKLEVEL", ARG_INT, sizeof(int), (void *)attacklevel);
+ return _plug_get_category(desc);
 }
 
 
+
 ExtFunc
 void plug_add_host(desc, hostname)
  struct arglist * desc;
@@ -417,15 +811,15 @@ void host_add_port_proto(args, portnum, state, proto)
  int state;
  char * proto;
 {
- char * port_s = emalloc(50+strlen(proto));
- sprintf(port_s, "Ports/%s/%d", proto, portnum);
+ char port_s[255];
+ 
+ snprintf(port_s, sizeof(port_s), "Ports/%s/%d", proto, portnum);
  if(!plug_get_key(args, port_s))
  {
   plug_set_key(args, port_s, ARG_INT, (void*)1);
-  sprintf(port_s, "/tmp/Ports/%s/%d", proto, portnum);
+  snprintf(port_s, sizeof(port_s), "/tmp/Ports/%s/%d", proto, portnum);
   plug_set_key(args, port_s, ARG_INT, (void*)1);
  }
- free(port_s);
 }
 
 
@@ -444,7 +838,7 @@ void host_add_port_udp(hostdata, portnum, state)
  int portnum;
  int state;
 {
- host_add_port_proto(hostdata, portnum, state, estrdup("udp"));
+ host_add_port_proto(hostdata, portnum, state, "udp");
 }
   
 
@@ -459,63 +853,76 @@ int port_in_ports(port, ports, s, e)
 }
  	
 
+
+static int
+unscanned_ports_as_closed(prefs)
+ struct arglist * prefs;
+{
+ char * unscanned;
+ unscanned = arg_get_value(prefs, "unscanned_closed");
+ if(unscanned && !strcmp(unscanned, "yes"))
+  return 0;
+ else
+  return 1;
+}
+           
 ExtFunc
-int host_get_port_state_proto(plugdata, portnum, proto)
- struct arglist * plugdata;
+int kb_get_port_state_proto(kb, prefs, portnum, proto)
+ struct arglist * kb;
+ struct arglist * prefs;
  int portnum;
  char * proto;
 { 
- char * port_s;
+ char port_s[255];
  unsigned short * range;
- int flag;
+ char * prange = (char*)arg_get_value(prefs, "port_range");
  int num;
 
-
  if(!proto)
   proto = "tcp";
   
  /* Check that we actually scanned the port */
  
- if(!strcmp(proto, "tcp") && !plug_get_key(plugdata, "Host/scanned")){
-#ifdef DEBUG
-	printf("HOST NOT SCANNED\n");
-#endif
-	return(1);
+ if(!strcmp(proto, "tcp") && arg_get_value(kb, "Host/scanned") == NULL){
+	return unscanned_ports_as_closed(prefs);
 	}
 
- else if(!strcmp(proto, "udp") && !plug_get_key(plugdata, "Host/udp_scanned"))
+ else if(!strcmp(proto, "udp") && arg_get_value(kb, "Host/udp_scanned") == NULL)
        {
         return 1;
       }
       
       		
- range = arg_get_value(plugdata, "ports");
- if(!range){
-#ifdef DEBUG
-	printf("unknown port range\n");
-#endif	
+ range = (u_short*)getpts(prange, &num);
+
+ if( range == NULL ){
  	return(1);
 	}
- num = (int)arg_get_value(plugdata, "ports_num");
+	
  if(!port_in_ports(portnum, range, 0, num)){
-#ifdef DEBUG
-	printf("!port_in_ports\n");
-#endif	
-       return(1);
+       return unscanned_ports_as_closed(prefs);
        }
  
-
  /* Ok, we scanned it. What is its state ? */
- port_s = emalloc(20+strlen(proto));
- sprintf(port_s, "Ports/%s/%d", proto, portnum);
- flag = (int)plug_get_key(plugdata, port_s);
- free(port_s);
-#ifdef DEBUG
- printf("%d -> %d\n", portnum, flag);
-#endif
- return(flag ? 1:0);
+ 
+ snprintf(port_s, sizeof(port_s), "Ports/%s/%d", proto, portnum);
+ if(arg_get_value(kb, port_s) == NULL)
+  return 0;
+ else
+  return 1;
 }
 
+ExtFunc
+int host_get_port_state_proto(plugdata, portnum, proto)
+ struct arglist * plugdata;
+ int portnum;
+ char * proto;
+{ 
+ struct arglist * kb = arg_get_value(plugdata, "key");
+ struct arglist * prefs = arg_get_value(plugdata, "preferences");
+ 
+ return kb_get_port_state_proto(kb, prefs, portnum, proto);
+}
 
 ExtFunc
 int host_get_port_state(plugdata, portnum)
@@ -537,6 +944,15 @@ int host_get_port_state_udp(plugdata, portnum)
 ExtFunc
 const char * plug_get_hostname(desc)
  struct arglist * desc;
+{
+ struct arglist * hinfos = arg_get_value(desc, "HOSTNAME");
+ if(hinfos)return((char*)arg_get_value(hinfos, "NAME"));
+ else return(NULL);
+}
+
+ExtFunc
+const char * plug_get_host_fqdn(desc)
+ struct arglist * desc;
 {
  struct arglist * hinfos = arg_get_value(desc, "HOSTNAME");
  if(hinfos)return((char*)arg_get_value(hinfos, "FQDN"));
@@ -559,20 +975,20 @@ static void
 mark_successful_plugin(desc)
  struct arglist * desc;
 {
- int id = (int)arg_get_value(desc, "ID");
+ int id = plug_get_id(desc);
  char data[512];
  
 
  bzero(data, sizeof(data));
- sprintf(data, "Success/%d", id);
- if(!plug_get_key(desc, data))
+ snprintf(data, sizeof(data), "Success/%d", id);
+ if( plug_get_key(desc, data) == NULL )
   plug_set_key(desc, data, ARG_INT, (void*)1);
  
  
  /*
   * KB entries starting by /tmp/ are not saved
   */
- sprintf(data, "/tmp/Success/%d", id);
+ snprintf(data, sizeof(data), "/tmp/Success/%d", id);
  if(!plug_get_key(desc, data))
   plug_set_key(desc, data, ARG_INT, (void*)1);
   
@@ -584,16 +1000,18 @@ mark_post(desc, action, content)
  char * action;
  char * content;
 {
- char * entry_name = emalloc(strlen(action) + 50);
+ char entry_name[255];
  int num_post = (int)arg_get_value(desc, "NUM_POST");
  
 #ifdef DEBUG_DIFF_SCAN
  printf("===>MARK_POST\n");
 #endif
- sprintf(entry_name, "SentData/%d/%s/%d", arg_get_value(desc, "ID"), action, num_post);
+
+ if(strlen(action) > (sizeof(entry_name) - 20))
+  return;
+  
+ snprintf(entry_name, sizeof(entry_name), "SentData/%d/%s/%d", plug_get_id(desc), action, num_post);
  plug_set_key(desc, entry_name, ARG_STRING, content);
- 
- efree(&entry_name);
 }
 
 static int
@@ -606,7 +1024,7 @@ post_sent_already(desc, action, content)
  int num_post = (int)arg_get_value(desc, "NUM_POST");
  struct arglist * key = (struct arglist *)arg_get_value(desc, "key");
  
- sprintf(trunc_name, "SentData/%d/%s/%d", arg_get_value(desc, "ID"), action,num_post);
+ snprintf(trunc_name, strlen(action) + 50, "SentData/%d/%s/%d", plug_get_id(desc), action,num_post);
  while(key && key->next)
  {
 #ifdef DEBUG_DIFF_SCAN
@@ -614,11 +1032,6 @@ post_sent_already(desc, action, content)
 #endif  
   if(!strcmp(trunc_name, key->name))
   {
-   char * v = key->value;
-   while((v = strchr(v, '')))v[0]='\n';
-   v = key->value;
-   while((v = strchr(v, '')))v[0]='\r';
-   
 #ifdef DEBUG_DIFF_SCAN
    printf("Compare %s and %s\n", trunc_name, key->name);
 #endif   
@@ -654,14 +1067,14 @@ proto_post_wrapped(desc, port, proto, action, what)
  char *t;
  char * buffer;
  int soc;
- struct arglist * globs;
- struct arglist * hostdata;
- char c;
  char * naction;
  int len;
  ntp_caps* caps = arg_get_value(desc, "NTP_CAPS");
  char * cve;
+ char * bid;
+ char * xref;
  int do_send = 1;
+ int i;
  int num_post = (int)arg_get_value(desc, "NUM_POST");
  
  if(!num_post)
@@ -678,66 +1091,104 @@ proto_post_wrapped(desc, port, proto, action, what)
  
  if(!action)
  {
-  action = arg_get_value(desc, "DESCRIPTION");
+  action = plug_get_description(desc);
  }
  
- cve = arg_get_value(desc, "CVE_ID");
- if(!action)return;
- len = strlen(action);
- if(cve)len+=strlen(cve)+20;
- c = (char)13;
+ cve = plug_get_cve_id(desc);
+ bid = plug_get_bugtraq_id(desc);
+ xref = plug_get_xref(desc);
  
+ if( action == NULL )
+ 	return;
+	
+	
+ len = strlen(action) + 1;
+ if(cve != NULL)
+ 	len += strlen(cve) + 20;
 
- if(!caps)return;
+ if(bid != NULL)
+  	len += strlen(bid) + 20;
+	
+ if(xref != NULL )
+ 	len += strlen(xref) + 20;
+	
+ if( caps == NULL )
+ 	return;
 
- hostdata = arg_get_value(desc, "HOSTNAME");
+ 
  naction = emalloc(len+1);
- if(cve)sprintf(naction, "%s\nCVE : %s\n", action, cve);
- else strncpy(naction, action, len);
- while((t=strchr(naction, '\n'))||(t=strchr(naction, c)))t[0]=';';
- buffer=malloc(1024+len);
+ strncpy(naction, action, strlen(action));
+ strcat(naction, "\n");
+ if( cve != NULL && cve[0] != '\0')
+        {
+	 strcat(naction, "CVE : ");
+	 strcat(naction, cve);
+	 strcat(naction, "\n");
+	 }
+ 
+ if( bid != NULL && bid[0] != '\0' )
+ 	{
+	 strcat(naction, "BID : ");
+	 strcat(naction, bid);
+	 strcat(naction, "\n");
+	 }	
+ if( xref != NULL && xref[0] != '\0' )
+ 	{
+	strcat(naction, "Other references : ");
+	strcat(naction, xref);
+	strcat(naction, "\n");
+	}
+ 
+ if( caps->escape_crlf == 0 )
+   while((t=strchr(naction, '\n'))||(t=strchr(naction, '\r')))t[0]=';';
+ else
+  {
+   char * old = naction;
+   len -= strlen(naction);
+   naction = addslashes(naction);
+   len += strlen(naction);
+   efree(&old);
+  }
+  
+ for(i=0;naction[i];i++)
+ {
+   if(!isprint(naction[i]))
+     	naction[i] = ' ';
+ }
+
+
+ buffer = emalloc( 1024 + len );
  if(caps->ntp_11) {
-   struct servent * service = NULL;
    char idbuffer[32];
-   service = getservbyport(htons((unsigned short)port), proto);
-#ifndef NESSUSNT
-   endservent();
-#endif
-   if (caps->scan_ids) {
-     /* XXX
-      * This check can later be removed when we can trust in the ID being
-      * present. It is mainly because I'm not sufficiently used to the
-      * Nessus sources.
-      *
-      * 29.01.1999, Jochen Wiedmann
-      */
-     if (arg_get_type(desc, "ID") == -1) {
+   char	*svc_name = (char*)nessus_get_svc_name(port, proto);
+   if (caps->scan_ids) { 
+     if (plug_get_id(desc) == 0) {
        *idbuffer = '\0';
      } else {
-       int id = (int)arg_get_value(desc, "ID");
-       sprintf(idbuffer, "<|> %d ", id);
+       int id = plug_get_id(desc);
+       snprintf(idbuffer, sizeof(idbuffer), "<|> %d ", id);
      }
    } else {
      *idbuffer = '\0';
  }
   if(port>0){
-     sprintf(buffer,
+     snprintf(buffer, 1024 + len, 
 	     "SERVER <|> %s <|> %s <|> %s (%d/%s) <|> %s %s<|> SERVER\n",
 	     what,
-  	(char *)arg_get_value(hostdata, "FQDN"),
-        service ? service->s_name:"unknown",
+  	     plug_get_hostname(desc),
+	     svc_name,
 	     port, proto, naction, idbuffer);
     
    } else
-     sprintf(buffer,
+     snprintf(buffer, 1024 + len, 
 	     "SERVER <|> %s <|> %s <|> general/%s <|> %s %s<|> SERVER\n",
 	     what,
-  	(char *)arg_get_value(hostdata, "FQDN"), 
+  	     plug_get_hostname(desc), 
 	     proto, naction, idbuffer);
  } else {
-   sprintf(buffer, "SERVER <|> %s <|> %s <|> %d:%s <|> SERVER\n", 
+   snprintf(buffer, 1024 + len, "SERVER <|> %s <|> %s <|> %d:%s <|> SERVER\n", 
 	   what,
-	   (char *)arg_get_value(hostdata, "FQDN"), port, naction);
+	   plug_get_hostname(desc), port, naction);
  }
  
  
@@ -749,11 +1200,9 @@ proto_post_wrapped(desc, port, proto, action, what)
   */
  if(arg_get_value(desc, "DIFF_SCAN"))
  {
-  char str[128];
 #ifdef DEBUG_DIFF_SCAN  
   fprintf(stderr, "** DIFF_SCAN enabled\n");
 #endif  
-  bzero(str, sizeof(str));
   
   if(!post_sent_already(desc, what, action))
  	 {
@@ -761,17 +1210,26 @@ proto_post_wrapped(desc, port, proto, action, what)
  	 }
   else do_send = 0;
  }
-#ifdef DEBUG_DIFF_SCAN
- else fprintf(stderr, "*** DIFF_SCAN : NO\n");
-#endif 
+  else
+    mark_post(desc, what, action);
+
  
- if(do_send)
+ if(do_send != 0)
  {
+  char ack;
+  int n = 0, e, l;
+  
   soc = (int)arg_get_value(desc, "SOCKET");
-  globs = emalloc(sizeof(struct arglist));
-  arg_add_value(globs, "global_socket", ARG_INT, sizeof(int), (void *)soc);
-  auth_send(globs, buffer);
-  arg_free_all(globs);
+  l = strlen(buffer);
+  while (n != l)
+  {
+   e = send(soc, buffer + n, l - n, 0);
+   if(e <= 0)
+    break;
+   else
+    n += e;
+  }
+  recv(soc, &ack, 1, 0);
  }
  
  /*
@@ -854,6 +1312,9 @@ void post_note(desc, port, action)
  int port;
  const char * action;
 {
+#if 0
+  fprintf(stderr, "Post_note: port = %d action = %s\n", port, action);
+#endif
   proto_post_note(desc, port, "tcp", action);
 } 
 
@@ -901,21 +1362,33 @@ char * get_preference(desc, name)
 
 
 ExtFunc
-void add_plugin_preference(desc, name, type, defaul)
- struct arglist *desc;
+void _add_plugin_preference(prefs, p_name, name, type, defaul)
+ struct arglist *prefs;
+ const char * p_name;
  const char * name;
  const char * type;
  const char * defaul;
 {
- struct arglist * prefs = arg_get_value(desc, "preferences");
- char * p_name = arg_get_value(desc, "NAME");
  char * pref;
  char * cname;
+ int len;
  
- cname = strdup(name);
- while(cname[strlen(cname)-1]==' ')cname[strlen(cname)-1]='\0';
  
- if(!prefs || !p_name)return;
+ 
+ cname = estrdup(name);
+ len = strlen(cname);
+ while(cname[len-1]==' ')
+ {
+  cname[len-1]='\0';
+  len --;
+ }
+ if(!prefs || !p_name)
+   {
+     efree(&cname);
+     return;
+   }
+
+
  pref = emalloc(strlen(p_name)+10+strlen(type)+strlen(cname));
  sprintf(pref, "%s[%s]:%s", p_name, type, cname);
  arg_add_value(prefs, pref, ARG_STRING, strlen(defaul), estrdup(defaul));
@@ -923,22 +1396,59 @@ void add_plugin_preference(desc, name, type, defaul)
  efree(&pref);
 }
 
+void add_plugin_preference(desc, name, type, defaul)
+ struct arglist *desc;
+ const char * name;
+ const char * type;
+ const char * defaul;
+{
+ struct arglist * prefs = arg_get_value(desc, "PLUGIN_PREFS");
+ char pref[1024];
+ 
+ 
+ if(prefs == NULL)
+  {
+   prefs = emalloc(sizeof(struct arglist));
+   arg_add_value(desc, "PLUGIN_PREFS", ARG_ARGLIST, -1, prefs);
+  }
+ 
 
-ExtFunc
-char * get_plugin_preference(desc, name)
+ snprintf(pref, sizeof(pref), "%s/%s", type, name);
+ arg_add_value(prefs, pref, ARG_STRING, strlen(defaul), estrdup(defaul));
+}
+
+
+
+ExtFunc char * 
+get_plugin_preference(desc, name)
   struct arglist * desc;
   const char * name;
 {
  struct arglist * prefs = arg_get_value(desc, "preferences");
- char * plug_name = arg_get_value(desc, "NAME");
- char * cname = strdup(name);
- while(cname[strlen(cname)-1]==' ')cname[strlen(cname)-1]='\0';
+ char * plug_name = plug_get_name(desc);
+ char * cname = estrdup(name);
+ int len;
  
- while(prefs && prefs->next)
+ len = strlen(cname);
+ 
+ while(cname[len-1]==' ')
+ {
+  cname[len-1]='\0';
+  len --;
+ }
+ 
+ 
+ if(!prefs)
+   {
+     efree(&cname);
+     return NULL;
+   }
+  
+ while(prefs->next)
  {
   char * a= NULL, *b = NULL;
   int c = 0;
-  char * t = estrdup(prefs->name);
+  char * t = prefs->name;
   
   a = strchr(t, '[');
   if(a)b=strchr(t, ']');
@@ -948,21 +1458,39 @@ char * get_plugin_preference(desc, name)
   {
    b+=2*sizeof(char);
    if(!strcmp(cname, b)){
+   	int old = a[0];
    	a[0] = 0;
 	if(!strcmp(t, plug_name)){
-		efree(&t);
+		a[0] = old;
 		efree(&cname);
 		return(prefs->value);
 		}
+	a[0] = old;	
 	}
   }
-  efree(&t);
   prefs = prefs->next;
  }
  efree(&cname);
  return(NULL);
 }
 
+ExtFunc const char * 
+get_plugin_preference_fname(desc, filename)
+ struct arglist * desc;
+ const char * filename;
+{
+ struct arglist * globals = arg_get_value(desc, "globals");
+ harglst * trans;
+ if(!globals) 
+  return NULL;
+  
+ trans = arg_get_value(globals, "files_translation");
+ if(!trans)
+  return NULL;
+ 
+ return harg_get_string(trans, filename);
+}
+
 
 ExtFunc
 void plug_set_key(args, name, type, value)
@@ -980,55 +1508,44 @@ void plug_set_key(args, name, type, value)
  if(!name || !value)return;
  switch(type)
  {
-   char * t;
   case ARG_STRING :
+   value = addslashes(value);
    str = emalloc(strlen(name)+strlen(value)+10);
-   t = value;
-   while((t = strchr(t, '\n')))t[0]='';
-   t = value;
-   while((t = strchr(t, '\r')))t[0]='';
-
    sprintf(str, "%d %s=%s;\n", ARG_STRING, name, (char *)value);
-   t = value;
-   while((t = strchr(t, '')))t[0]='\n';
-   t = value;
-   while((t = strchr(t, '')))t[0]='\r';
+   efree(&value);
    break;
   case ARG_INT :
    str = emalloc(strlen(name)+20);
    sprintf(str, "%d %s=%d;\n", ARG_STRING, name, (int)value);
    break;
-#ifdef NOT_IMPLEMENTED_YET
-  case ARG_ARGLIST :
-   {
-    struct arglist * t = (struct arglist *)value;
-    str = emalloc(20+strlen(name));
-    sprintf(str, "%d %s/START_AL\n", type, name);
-    write(pip, str, strlen(str));
-    while(t && t->next){
-    	plug_set_key(args, t->name, t->type, t->value);
-        t = t->next;
-        }
-    sprintf(str, "%d %s/END_AL\n",type, name);
-   }
-   break;
-#endif
  }
  if(str)
  {
    int len = strlen(str);
    int sent = 0;
+
+   /*
+    * Don't send a string larger than 65k
+    */
+   if( len > 65535 )
+   {
+     str[65533] = ';';
+     str[65534] = '\n';
+     str[65535] = '\0';
+   }
+
    
    while(sent < len)
    {
-     int e;
-    e = send(pip, str+sent, len+1, 0);
-    if(e < 0){
-    	perror("send ");
-	return;
+    int e;
+    e = send(pip, str+sent, len-sent, 0);
+    if(e <= 0){
+        fprintf(stderr, "[%d] plug_set_key:send(%d)['%s'](%d out of %d): %s\n",getpid(), pip,str, sent, len, strerror(errno));
+	break;
 	}
     sent+=e;
    }
+   efree(&str);
   }
 } 
 
@@ -1041,44 +1558,44 @@ scanner_add_port(args, port, proto)
  char * proto;
 {
  ntp_caps* caps = arg_get_value(args, "NTP_CAPS");
- struct arglist * hostdata = arg_get_value(args, "HOSTNAME");
  char * buf;
- struct servent * serv = getservbyport(htons(port), proto);
- char * hn = arg_get_value(hostdata, "FQDN");
+ char *svc_name = (char*)nessus_get_svc_name(port, proto);
+ char * hn = (char*)plug_get_hostname(args);
  int len;
  int soc;
  struct arglist * globs;
  int do_send = 1;
-#ifndef NESSUSNT
-  endservent();
-#endif 
- 
+ static int confirm = -1;
  
+ if(confirm < 0)
+ {
+  struct arglist * globals = arg_get_value(args, "globals");
+  if(globals)confirm = (int)arg_get_value(globals, "confirm");
+ }
+
  /*
   * Diff scan stuff : if the port was known to be open,
   * there is no need to report it again.
   */
  if(arg_get_value(args, "DIFF_SCAN"))
  {
-   char * port_s = emalloc(50+strlen(proto));
-   sprintf(port_s, "Ports/%s/%d", proto, port);
+   char port_s[255];
+   snprintf(port_s, sizeof(port_s), "Ports/%s/%d", proto, port);
    if(plug_get_key(args, port_s))do_send = 0;
-   efree(&port_s);
  }
- 
+
+
  host_add_port_proto(args, port, 1, proto);
  
- len = 255 + (hn ? strlen(hn):0) + 
- 	    (serv ?strlen(serv->s_name):strlen("unknown"));
+ len = 255 + (hn ? strlen(hn):0) + strlen(svc_name);
  buf = emalloc(len);
- if(caps->ntp_11)
-  sprintf(buf, "SERVER <|> PORT <|> %s <|> %s (%d/%s) <|> SERVER\n",
- 		hn,serv ? serv->s_name:"unknown",
-		port, proto);
+ if(caps != NULL && caps->ntp_11)
+  snprintf(buf, len, "SERVER <|> PORT <|> %s <|> %s (%d/%s) <|> SERVER\n",
+ 		hn,svc_name, port, proto);
  else
   {
    if(!strcmp(proto, "tcp"))
-     sprintf(buf, "SERVER <|> PORT <|> %s <|> %d <|> SERVER\n",
+     snprintf(buf, len, "SERVER <|> PORT <|> %s <|> %d <|> SERVER\n",
   		hn, port);
   }
    
@@ -1087,11 +1604,12 @@ scanner_add_port(args, port, proto)
  soc = (int)arg_get_value(args, "SOCKET");
  globs = emalloc(sizeof(struct arglist));
  arg_add_value(globs, "global_socket", ARG_INT, sizeof(int), (void *)soc);
+ arg_add_value(globs, "confirm", ARG_INT, sizeof(int), (void*)confirm);
  
  auth_send(globs, buf);
  arg_free(globs);
  }
- free(buf);
+ efree(&buf);
 }
 
 
@@ -1101,28 +1619,72 @@ scanner_add_port(args, port, proto)
  * itself
  */
 #ifndef NESSUSNT
-static int _plug_get_key_son;
+static int _plug_get_key_son = 0;
 
 static void 
 plug_get_key_sighand_term(int sig)
 {
- if(_plug_get_key_son)
+ int son = _plug_get_key_son;
+ 
+ if(son != 0)
  {
-  kill(_plug_get_key_son, SIGTERM);
+  kill(son, SIGTERM);
   _plug_get_key_son = 0;
  }
- exit(0);
+ _exit(0);
+}
+
+static void
+plug_get_key_sigchld(int sig)
+{
+ int status;
+ wait(&status);
+}
+
+static void
+sig_n(int signo, void (*fnc)(int) )
+{
+ #ifdef HAVE_SIGACTION
+  struct sigaction sa;
+  sa.sa_handler = fnc;
+  sa.sa_flags = 0;
+  sigemptyset(&sa.sa_mask);
+  sigaction(signo, &sa, (struct sigaction *) 0);
+#else
+  signal(signo, fnc);
+#endif
+}
+
+static void
+sig_term( void (*fcn)(int) )
+{
+ sig_n(SIGTERM, fcn);
+}
+
+static void
+sig_alarm( void (*fcn)(int) )
+{
+ sig_n(SIGALRM, fcn);
+}
+
+static void 
+sig_chld( void(*fcn)(int) )
+{
+ sig_n(SIGCHLD, fcn);
 }
 #endif
 
-void * plug_get_key(args, name)
+
+void * 
+plug_get_key(args, name)
  struct arglist * args;
  char * name;
 {
-#ifndef NESSUSNT
  struct arglist * key = (struct arglist *)arg_get_value(args, "key");
  int type;
  
+ 
+ 
  if(!key)return(NULL);
  
  type = arg_get_type(key, name);
@@ -1130,49 +1692,61 @@ void * plug_get_key(args, name)
  {
   if(type == ARG_STRING)
      {
-      char * t = arg_get_value(key, name);
-      char * v = t;
-      if(v)
-      {
-       while((v = strchr(v, '')))v[0]='\n';
-       v = t;
-       while((v = strchr(v, '')))v[0]='\r';
-      }
-      
-      return t;
+      return arg_get_value(key, name);
      }
    else if(type == ARG_INT)
     return arg_get_value(key,name);
   else if(type == ARG_ARGLIST)
   {
    struct arglist * value = arg_get_value(key, name);
+   sig_chld(plug_get_key_sigchld);
    while(value && value->next)
    {
      int pid;  
-    if(!(pid = fork()))
+     if(!(pid = fork()))
      {
+     /* so that different plugins do not use the same seed */  
+     srand48(getpid() + getppid() + time(NULL));
+ 
+     sig_term(_exit);
+     sig_alarm(_exit);
+     alarm(120);
      arg_set_value(key, name, -1, value->value);
      arg_set_type(key, name, value->type);
      return value->value;
      }
     else
      {
-       void* old;
+      if(pid < 0)
+      {
+       fprintf(stderr, "nessus-libraries:libnessus:plugutils.c:plug_get_key(): fork() failed : %s", strerror(errno));	      
+       return NULL;
+      }
+      else
+      {
+      int e;
+      int status;
       _plug_get_key_son = pid;
-      old = signal(SIGTERM, plug_get_key_sighand_term);
-      waitpid(pid,NULL, 0);
-      signal(SIGTERM, old);
+      sig_term(plug_get_key_sighand_term);
+      for(;;)
+      {
+      e = waitpid(pid,&status, 0);
+      if( e < 0 && errno == EINTR)
+         continue; 
+      else
+        if(e < 0 || WIFEXITED(status) != 0 || WIFSIGNALED(status) != 0)
+		break;
+      }
       _plug_get_key_son = 0;
+      sig_term(_exit);
+      }
     }
     value = value->next;
    }
    exit(0);
   }
  }
- else return NULL;
-#else
  return NULL;
-#endif  
 }
 
 ExtFunc unsigned int 
@@ -1180,22 +1754,218 @@ plug_get_host_open_port(struct arglist * desc)
 {
  struct arglist * h = arg_get_value(desc, "key");
  char * str = "Ports/tcp/";
+ int    str_len = strlen(str);
  char * t;
  int port = 0;
+ u_short candidates[64];
+ int num_candidates = 0;
+ int open21 = 0, open80 = 0;
  
  while(h && h->next)
  {
-  if((strlen(h->name) > strlen(str)) && 
-     !strncmp(h->name, str, strlen(str))){
+  if( ( strlen(h->name) > str_len ) && 
+      ( strncmp(h->name, str, str_len ) == 0 ) ){
      	t = h->name + strlen(str);
 	port = atoi(t);
-	break;
+	/* 
+	 * Transparent proxies may run on these ports, so
+	 * we try to avoid them
+	 */
+	if((port != 21) && (port != 80)) {
+		candidates[num_candidates++] = port;
+		if ( num_candidates >= 64 ) break;
+		}
+	else 
+		{
+		if ( port == 21 ) open21 = 1;
+		else if ( port == 80 ) open80 = 1;
+		}
 	}
   h = h->next;	
  }
- 
- return(port);
+
+ if ( num_candidates != 0 )
+ {
+  return candidates[lrand48() % num_candidates];
+ }
+ else if (open21) return 21;
+ else if (open80) return 80;
+ else return 0;
 }
        
  
+/*
+ * Those brain damaged functions should probably be in another file
+ * They are use to remember who speaks SSL or not
+ */
+   
+ExtFunc
+void plug_set_port_transport(args, port, tr)
+     struct arglist * args;
+     int		port, tr;
+{
+  static char	s[256];
 
+  snprintf(s, sizeof(s), "Transports/TCP/%d", port);
+  plug_set_key(args, s, ARG_INT, (void*) tr);
+}
+
+ExtFunc
+int plug_get_port_transport(args, port)
+     struct arglist * args;
+     int		port;
+{
+  static char	s[256];
+  char * trp;
+  snprintf(s, sizeof(s), "Transports/TCP/%d", port);
+  
+  trp = plug_get_key(args, s);
+  if(trp)
+   return atoi(trp);
+  else
+   return NESSUS_ENCAPS_IP;
+   /* 
+    * Change the above to '0' to make ultra-smart
+    * SSL negotation which may crash stuff
+    */
+}
+
+ExtFunc
+const char* plug_get_port_transport_name(args, port)
+     struct arglist * args;
+     int		port;
+{
+  return get_encaps_name(plug_get_port_transport(args, port));
+}
+
+#ifdef HAVE_SSL
+static void
+plug_set_ssl_item(args, item, itemfname)
+ struct arglist * args;
+ char * item;
+ char * itemfname;
+{
+ static char s[256];
+ struct arglist * key = (struct arglist*)arg_get_value(args, "key");
+ snprintf(s, sizeof(s), "SSL/%s", item);
+ plug_set_key(args, s, ARG_STRING, itemfname);
+
+ /*
+  * Ugly hack - the KB is not updated before the next plugin call, 
+  * so we manually add this key
+  */
+ arg_add_value(key, s, ARG_STRING, strlen(itemfname), estrdup(itemfname));
+}
+
+ExtFunc void
+plug_set_ssl_cert(args, cert)
+ struct arglist * args;
+ char * cert;
+{
+ plug_set_ssl_item(args, "cert", cert);
+}
+
+ExtFunc void 
+plug_set_ssl_key(args, key)
+ struct arglist * args;
+ char * key;
+{
+ plug_set_ssl_item(args, "key", key);
+}
+ExtFunc void
+plug_set_ssl_pem_password(args, key)
+ struct arglist * args;
+ char * key;
+{
+ plug_set_ssl_item(args, "password", key);
+}
+
+ExtFunc void
+plug_set_ssl_CA_file(args, key)
+ struct arglist * args;
+ char * key;
+{
+ plug_set_ssl_item(args, "CA", key);
+}
+#else
+ExtFunc  void
+plug_set_ssl_cert(args, cert)
+ struct arglist * args;
+ char * cert;
+{
+ fprintf(stderr, "plug_set_ssl_cert(): not implemented\n");
+}
+
+ExtFunc void
+plug_set_ssl_key(args, key)
+ struct arglist * args;
+ char * key;
+{
+ fprintf(stderr, "plug_set_ssl_key(): not implemented\n");
+}
+#endif /* HAVE_SSL */
+
+
+ExtFunc char *
+find_in_path(name, safe)
+     char	*name;
+     int	safe;
+{
+  char		*buf = getenv("PATH"), *pbuf, *p1, *p2;
+  static char	cmd[MAXPATHLEN];
+  int		len = strlen(name);
+  
+  if (len >= MAXPATHLEN)
+    return NULL;
+
+#if 0
+  /* Proposed by Devin Kowatch 
+     If it's already an absolute path take it as is */
+  if (name[0] == '/' && access(name, X_OK) == 0)
+    return name;
+#endif
+
+  if (buf == NULL)		/* Should we use a standard PATH here? */
+    return NULL;
+
+  pbuf = buf;
+  while (*pbuf != '\0')
+    {
+      for (p1 = pbuf, p2 = cmd; *p1 != ':' && *p1 != '\0'; )
+	*p2 ++ = *p1 ++;
+      *p2 = '\0';
+      if (*p1 == ':')
+	p1 ++;
+      pbuf = p1;
+      if (p2 == cmd)		/* :: found in $PATH */
+	strcpy(cmd, ".");
+
+      if (cmd[0] != '/' && safe)
+	continue;
+      if (p2 - cmd + 1 + len >= MAXPATHLEN)
+	/* path too long: cannot be reached */
+	continue;
+
+      sprintf(p2, "/%s", name);
+      if (access(cmd, X_OK) == 0)
+	{
+	  *p2 = '\0';
+#if 0
+	  fprintf(stderr, "find_in_path: %s found in %s\n", name, cmd);
+#endif
+	  return cmd;
+	}
+#if 0
+      else
+	fprintf(stderr, "find_in_path: No %s\n", cmd);
+#endif
+    }
+  return NULL;
+}
+
+ExtFunc int 
+is_shell_command_present(name)
+ char * name;
+{
+  return find_in_path(name, 0) != NULL;
+}
