@@ -21,12 +21,13 @@
 #define EXPORTING
 #include <includes.h>
 #include "comm.h"
+#include "services.h"
  
 /*
  * Sends the status of an action
  */
 ExtFunc
-void 
+int 
 comm_send_status(globals, hostname, action, current,max)
   struct arglist * globals;
   char * hostname;
@@ -37,6 +38,17 @@ comm_send_status(globals, hostname, action, current,max)
  char * pref = arg_get_value(prefs, "ntp_short_status");
  int short_status;
  ntp_caps* caps = arg_get_value(globals, "ntp_caps");
+ int soc = (int)arg_get_value(globals, "global_socket");
+ char buffer[2048];
+ int e, n = 0, l, ack;
+ 
+ 
+ if (soc < 0 || soc > 1024) 
+  return -1;
+ 
+ 
+ if(strlen(hostname) > (sizeof(buffer) - 50))
+  return -1;
   
  if(pref && !strcmp(pref, "yes"))
   short_status = 1;
@@ -47,19 +59,40 @@ comm_send_status(globals, hostname, action, current,max)
     {
     if(short_status)
       {
-      char *b = emalloc(strlen(hostname)+50);
-      sprintf(b, "s:%c:%s:%d:%d\n", action[0], hostname, current, max);
-      auth_send(globals, b);
-      efree(&b);
+      snprintf(buffer, sizeof(buffer), "s:%c:%s:%d:%d\n", action[0], hostname, current, max);
       }
     else
-    auth_printf(globals,
+     snprintf(buffer, sizeof(buffer),
 		"SERVER <|> STATUS <|> %s <|> %s <|> %d/%d <|> SERVER\n",
 		hostname, action, current, max);
     }
   else
-   auth_printf(globals, "SERVER <|> STAT <|> %s <|> %d/%d <|> SERVER\n",
+   snprintf(buffer, sizeof(buffer), "SERVER <|> STAT <|> %s <|> %d/%d <|> SERVER\n",
 	      hostname, current, max);
+	      
+	      
+  /*
+   * Send our data
+   */	      
+  l = strlen(buffer);
+  while ( n != l )
+  {
+   e = send(soc, buffer + n , l - n , 0);
+   if( e < 0 && errno == EINTR )
+   	continue;
+	
+   if ( e <= 0 )
+    return -1;
+   else
+    n += e;
+  }
+  
+ do 
+ { 
+  e = recv(soc, &ack, 1, 0);
+ } while (e < 0 && errno == EINTR);
+ 
+ return 0;
 }
 
 
@@ -89,18 +122,60 @@ int qsort_compar(const void* a, const void* b)
  
 /* Convert a string like "-100,200-1024,3000-4000,60000-" into an array 
    of port numbers*/
-unsigned short *getpts(char *origexpr) {
-int exlen = strlen(origexpr);
+unsigned short *getpts(char *origexpr, int * len) {
+int exlen;
 char *p,*q;
 unsigned short *tmp, *ports;
 int i=0, j=0,start,end;
-char *expr = estrdup(origexpr);
-char *mem = expr;
+char *expr;
+char *mem;
+char * s_start, * s_end;
+static unsigned short * last_ret = NULL;
+static char * last_expr = NULL;
+static int last_num;
+
+ if(strcmp(origexpr, "default") == 0)
+	 return get_tcp_svcs(len);
+
+ expr = estrdup(origexpr);
+ exlen = strlen(origexpr);
+ mem = expr;
+
+if( last_expr != NULL)
+{
+ if(strcmp(last_expr, expr) == 0)
+  {
+  if(len != NULL)*len = last_num;
+  return last_ret;
+ }
+ else
+ { 
+  efree(&last_expr);
+  efree(&last_ret);
+ }
+}
+
+
+
 
 ports = emalloc(65536 * sizeof(short));
 for(;j < exlen; j++) 
   if (expr[j] != ' ') expr[i++] = expr[j]; 
 expr[i] = '\0';
+
+if((s_start = strstr(expr, "T:")) != NULL)
+{
+ expr = &(s_start[2]);
+}
+
+if((s_end = strstr(expr, "U:")) != NULL)
+{
+   if(s_end[-1] == ',')
+     s_end --;
+ s_end[0] = '\0';
+}
+
+
 exlen = i;
 i=0;
 while((p = strchr(expr,','))) {
@@ -134,13 +209,14 @@ ports[i++] = 0;
   
   
 
-#ifndef NESSUSNT /* WindowsNT doesn't like realloc() */
+
   qsort(ports, i, sizeof(u_short), qsort_compar);
   tmp = realloc(ports, i * sizeof(short));
+  if(len != NULL)*len = i - 1;
   efree(&mem);
+  
+  last_ret = tmp;
+  last_expr = estrdup(origexpr);
+  last_num =  i - 1;
   return tmp;
-#else
-  tmp = NULL;
-  return(ports);
-#endif
 }
